{************************************************************}
{              CADHandle Importer SDK DLL Version demo             }
{                                                            }
{      Copyright (c) 2002-2005 SoftGold software company     }
{                                                            }
{************************************************************}
{$DEFINE DRAWSPLINE}

unit Main;

interface

{$DEFINE SG_CADIMPORTERDLLDEMO}

//{$DEFINE MEMCHK}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, ComCtrls, StdCtrls, Math, Menus, Buttons, sgConsts,
  {$IFDEF MEMCHK}
  MemCheck,
  {$ENDIF}
  SGLines;

type
  TfmCADDLLdemo = class(TForm)
    pnlOptions: TPanel;
    PageControl1: TPageControl;
    tsDrawing: TTabSheet;
    pbDrawing: TPaintBox;
    OpenDialog1: TOpenDialog;
    MainMenu1: TMainMenu;
    mmiHome: TMenuItem;
    Exit1: TMenuItem;
    N1: TMenuItem;
    mmiOpen: TMenuItem;
    mmiPlaceImageToHome: TMenuItem;
    lblMeasure: TLabel;
    cbScale: TComboBox;                 
    cbSplitArcs: TCheckBox;
    cbLayers: TComboBox;
    pmiAbout: TMenuItem;
    pmiShowAbout: TMenuItem;
    cbUseWinLine: TCheckBox;
    cbDimensions: TCheckBox;
    sbAllLayers: TCheckBox;
    mmiZoomAll: TMenuItem;
    cbProhibitArcsAsCurves: TCheckBox;
    btnPlaceImageToHome: TSpeedButton;
    cbLayouts: TComboBox;
    cbTextAsCurves: TCheckBox;
    procedure pbDrawingPaint(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure cbScaleChange(Sender: TObject);
    procedure pbDrawingMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pbDrawingMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure pbDrawingMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure cbSplitArcsClick(Sender: TObject);
    procedure cbLayersDrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure mmiOpenClick(Sender: TObject);
    procedure mmiPlaceImageToHomeClick(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
    procedure pmiShowAboutClick(Sender: TObject);
    procedure mmiZoomAllClick(Sender: TObject);
    procedure cbProhibitArcsAsCurvesClick(Sender: TObject);
    procedure cbLayoutsChange(Sender: TObject);
    procedure cbUseWinLineClick(Sender: TObject);
    procedure cbTextAsCurvesClick(Sender: TObject);
  private
    FCurrentDir: string;
    FX: Integer;
    FY: Integer;
    FPlaceToHome: Boolean;
    FZoomAll: Boolean;
    FScale: Integer;
    FStart: TPoint;
    FOld: TPoint;
  public
    { Public declarations }
  end;

var
  fmCADDLLdemo: TfmCADDLLdemo;
  Lines: TsgLines;
  CADHandle: THandle;
  LScale: Double;

implementation

uses CADIntf;

{$R *.DFM}

type
  TMyPaint = class(TPaintBox)
  end;

var
  bTextByCurves: Boolean;
  EData: PcadData;
  IsInsideInsert: Boolean = False;

procedure Error;
var Buf: array[Byte] of Char;
begin
  CADGetLastError(Buf);
  raise Exception.Create(Buf);
end;

{ DrawPolyPolyLine

  Draws a polyline on points of the prepared list
  (list consists of PPoint-points).                 }
procedure DrawPolyPolyLine(DC: HDC; IntPoints: TList);
var
  Points: TList;
  I, N, C: Integer;
  P: Pointer;
  Pt1, Pt2, PtPrev: TPoint;
  isPrevDot: Boolean;
begin
  if DC = 0 then Exit;
  Points := TList.Create;
  try
    isPrevDot:= False;
    I := 0;
    while I < IntPoints.Count -1 do
    begin
      Pt1 :=  PPoint(IntPoints.Items[I])^;
      Pt2 :=  PPoint(IntPoints.Items[I+1])^;
      // For correct drawing and scaling dots
      if (Pt2.X = Pt1.X) and (Pt2.Y = Pt1.Y) then
      begin
        // If the current dot equally with previous dot
        if isPrevDot and (PtPrev.X = Pt1.X) and (PtPrev.Y = Pt1.Y) then
        begin
          Inc(I, 2);
          Continue;// miss of identical points
        end;
        PtPrev.X := Pt1.X;
        PtPrev.Y := Pt1.Y;
        Pt2.Y := Pt2.Y + 1;
        isPrevDot:= True;
      end
      else
        isPrevDot:= False;
      Points.Add(Pointer(Pt1.X));
      Points.Add(Pointer(Pt1.Y));
      Points.Add(Pointer(Pt2.X));
      Points.Add(Pointer(Pt2.Y));
      Inc(I, 2);
    end;
    N := Points.Count shr 2;
    if N = 0 then Exit;
    C := Points.Count;
    for I := 0 to N - 1 do
      Points.Add(Pointer(2));
    P := Points.List;
    Inc(PInteger(P), C);
    PolyPolyline(DC, Points.List^, P^, N);
  finally
    Points.Free;
  end;
end;

{ DrawPolyLine

  Draws a only solid polyline on points of the prepared list
  (list consists of PPoint-points).                 }
procedure DrawPolyLine(DC: HDC; IntPoints: TList);
var
  Points: TList;
  I: Integer;
  Pt, PtPrev: TPoint;
begin
  if DC = 0 then Exit;
  Points := TList.Create;
  try
    for I := 0 to IntPoints.Count - 1 do
    begin
      Pt := PPoint(IntPoints.Items[I])^;
      if (I <> 0) and (PtPrev.X = Pt.X) and (PtPrev.Y = Pt.Y) then
        Continue;// miss of identical points
      Points.Add(Pointer(Pt.X));
      Points.Add(Pointer(Pt.Y));
      PtPrev.X := Pt.X;
      PtPrev.Y := Pt.Y;
    end;
    Windows.Polyline(DC, Points.List^, Points.Count shr 1);
  finally
    Points.Free;
  end;
end;

{ GetPoint

  Converts CADHandle units to pixels.}
function GetPoint(const P: TFPoint): TPoint;
begin
  Result.X := fmCADDLLdemo.FX + Round(P.X * fmCADDLLdemo.FScale / 100);
  Result.Y := fmCADDLLdemo.FY - Round(P.Y * fmCADDLLdemo.FScale / 100);
end;

function PCanvas: TCanvas;
begin
  Result := fmCADDLLdemo.pbDrawing.Canvas;
end;

{ SetRotation

  Set font for rotated text.
procedure SetRotation;
var
  LF: TLogFont;
begin
  GetObject(PCanvas.Font.Handle, SizeOf(LF), @LF);
  LF.lfEscapement := Round(EData.Rotation * 10);	// in tenths of degrees
  while LF.lfEscapement < 0
  do Inc(LF.lfEscapement, 3600);
  LF.lfOrientation := LF.lfEscapement;
  PCanvas.Font.Handle := CreateFontIndirect(LF);
end;
}

{ GetPointsList

  Function converts the PFPoint-list in the PPoint-list
  with converting CADHandle units to pixels.                       }
procedure GetPointsList(SingPoints, IntPts: TList);
var
  I: Integer;
  DXFPt: TFPoint;
  PIntPt: PPoint;
begin
  for I := 0 to SingPoints.Count - 1 do
  begin
    New(PIntPt);
    DXFPt.X := (PFPoint(SingPoints.Items[I]))^.X;
    DXFPt.Y := (PFPoint(SingPoints.Items[I]))^.Y;
    DXFPt.Z := (PFPoint(SingPoints.Items[I]))^.Z;// for future version
    PIntPt.X := GetPoint(DXFPt).X;
    PIntPt.Y := GetPoint(DXFPt).Y;
    IntPts.Add(PIntPt);
  end;
end;

procedure DrawPoint;
var
  P: TPoint;
begin
  P := GetPoint(EData.Point);// EData.Point - point position
  SetPixelV(PCanvas.Handle, P.X, P.Y, PCanvas.Pen.Color);
end;

{ DrawGlobal

  Draws a (poly)line in global coordinates. }
procedure DrawGlobal(SPoints: TList; IsSolid: Boolean);
var
  I, N, PrevX, PrevY: Integer;
  P: TPoint;
  PDXF: TFPoint;
begin
  if SPoints.Count = 0 then
    Exit// Error
  else
  begin
    N := 2;
    if IsSolid then
      N := 1;
    I := 0;
    while I < SPoints.Count -1 do
    begin
      PDXF.X := PFPoint(SPoints.Items[I])^.X;
      PDXF.Y := PFPoint(SPoints.Items[I])^.Y;
      PDXF.Z := PFPoint(SPoints.Items[I])^.Z;// for future version
      P := GetPoint(PDXF);
      PrevX := P.X;
      PrevY := P.Y;
      PCanvas.MoveTo(P.X, P.Y);
      PDXF.X := PFPoint(SPoints.Items[I+1])^.X;
      PDXF.Y := PFPoint(SPoints.Items[I+1])^.Y;
      PDXF.Z := PFPoint(SPoints.Items[I+1])^.Z;// for future version
      P := GetPoint(PDXF);
      // For correct drawing and scaling dots
      if (P.X = PrevX) and (P.Y = PrevY) then
        PCanvas.LineTo(P.X, P.Y + 1)
      else
        PCanvas.LineTo(P.X, P.Y);
      Inc(I, N);
    end;
  end;
end;

{ DrawLine

  Draws a line.
  EData.Point - starting point
  EData.Point1 - ending point. }
procedure DrawLine;
var
  I: Integer;
  SingPoints, IntPoints, DottedSingPoints: TList;
  PSinglePt: PFPoint;
begin
  SingPoints := TList.Create;
  DottedSingPoints := TList.Create;
  New(PSinglePt);
  PSinglePt^.X := EData.Point.X;
  PSinglePt^.Y := EData.Point.Y;
  PSinglePt^.Z := EData.Point.Z;
  SingPoints.Add(PSinglePt);
  New(PSinglePt);
  PSinglePt^.X := EData.Point1.X;
  PSinglePt^.Y := EData.Point1.Y;
  PSinglePt^.Z := EData.Point1.Z;
  SingPoints.Add(PSinglePt);

  if fmCADDLLdemo.cbUseWinLine.Checked then // Global coordinates
  begin
    if Lines.IsSolid then
      DrawGlobal(SingPoints, Lines.IsSolid)
    else
    begin
      Lines.Line(TFPoint(SingPoints[0]^), TFPoint(SingPoints[1]^), DottedSingPoints);
      DrawGlobal(DottedSingPoints, Lines.IsSolid);
    end;
  end
  else
  begin
    IntPoints := TList.Create;
    if Lines.IsSolid then
    begin
      GetPointsList(SingPoints, IntPoints);
      DrawPolyLine(PCanvas.Handle, IntPoints);
    end
    else
    begin
      Lines.Line(TFPoint(SingPoints[0]^), TFPoint(SingPoints[1]^), DottedSingPoints);
      GetPointsList(DottedSingPoints, IntPoints);
      DrawPolyPolyLine(PCanvas.Handle, IntPoints);
    end;
    for I := 0 to IntPoints.Count - 1 do
      Dispose(IntPoints[I]);
    IntPoints.Free;
  end;

  // Clearing memory
  for I := 0 to DottedSingPoints.Count - 1 do
    Dispose(DottedSingPoints[I]);
  for I := 0 to SingPoints.Count - 1 do
    Dispose(SingPoints[I]);
  DottedSingPoints.Free;
  SingPoints.Free;
end;

{ DrawSolid

  4 points in EData
  The order in CADHandle file is 0-1-3-2. }
procedure DrawSolid;
var
  FSolid: array[0..3] of TPoint;
begin
  PCanvas.Pen.Width := 0;
  PCanvas.Brush.Style := bsSolid;
  PCanvas.Brush.Color := PCanvas.Pen.Color;
  FSolid[0] := GetPoint(EData.Point);
  FSolid[1] := GetPoint(EData.Point1);
  FSolid[2] := GetPoint(EData.Point3);
  FSolid[3] := GetPoint(EData.Point2);
  Polygon(PCanvas.Handle, FSolid[0], 4);
  PCanvas.Brush.Style := bsClear;
end;

procedure Draw3DFace;

  procedure Edge(const P1, P2: TFPoint);
  var
    Pt: array[0..1] of TPoint;
  begin
    Pt[0] := GetPoint(P1);
    Pt[1] := GetPoint(P2);
    Polygon(PCanvas.Handle, Pt[0], 2);
  end;

begin
  PCanvas.Pen.Width := 0;
  PCanvas.Brush.Style := bsSolid;
  PCanvas.Brush.Color := PCanvas.Pen.Color;
  if EData.Flags and 1 = 0 then
    Edge(EData.Point, EData.Point1);
  if EData.Flags and 2 = 0 then
    Edge(EData.Point1, EData.Point2);
  if EData.Flags and 4 = 0 then
    Edge(EData.Point2, EData.Point3);
  if EData.Flags and 8 = 0 then
    Edge(EData.Point3, EData.Point);
  PCanvas.Brush.Style := bsClear;
end;

{ DrawPoly

  Draws a (poly)line in global coordinates.
  EData.Count - number of polyline vertices
  EData.Points - pointer to point array.    }
procedure DrawPoly;
var
  I: Integer;
  P: PFPoint;
  PSinglePt: PFPoint;
  SingPoints, IntPoints, DottedSingPoints: TList;
begin
  SingPoints := TList.Create;
  DottedSingPoints := TList.Create;

  P := EData.Points;
  for I := 0 to EData.Count - 1 do
  begin
    //if I = 0 then PCanvas.MoveTo(GetPoint(P^).X, GetPoint(P^).Y);
    New(PSinglePt);
    PSinglePt^.X := P^.X;
    PSinglePt^.Y := P^.Y;
    PSinglePt^.Z := P^.Z;
    SingPoints.Add(PSinglePt);
    Inc(P);
  end;

  if fmCADDLLdemo.cbUseWinLine.Checked then // Global coordinates
  begin
    if Lines.IsSolid then
      DrawGlobal(SingPoints, Lines.IsSolid)
    else
    begin
      Lines.Poly(SingPoints, DottedSingPoints, False);
      DrawGlobal(DottedSingPoints, Lines.IsSolid);
    end;
  end
  else
  begin
    IntPoints := TList.Create;
    if Lines.IsSolid then
    begin
      GetPointsList(SingPoints, IntPoints);
      DrawPolyLine(PCanvas.Handle, IntPoints);
    end
    else
    begin
      Lines.Poly(SingPoints, DottedSingPoints, False);
      GetPointsList(DottedSingPoints, IntPoints);
      DrawPolyPolyLine(PCanvas.Handle, IntPoints);
    end;
    for I := 0 to IntPoints.Count - 1 do
      Dispose(IntPoints[I]);
    IntPoints.Free;
  end;

  // Clearing memory
  for I := 0 to DottedSingPoints.Count - 1 do
    Dispose(DottedSingPoints[I]);
  for I := 0 to SingPoints.Count - 1 do
    Dispose(SingPoints[I]);
  DottedSingPoints.Free;
  SingPoints.Free;
end;

{ GetSpline

  Gets a spline.  }
function GetSpline(ControlPoints, FitPoints, Knots: TList; ChordLen: Single): TList;

  function Normalize(N, I: Integer; Param: Double): Double;
  var
    V1, D1, V2, D2: Double;
  begin
    if N = 0 then
    begin
      if (Single(Knots[I]) <= Param) and (Param < Single(Knots[I + 1])) then
        Result := 1
      else
        Result := 0;
    end
    else
    begin
      D1 := Single(Knots[I + N]) - Single(Knots[I]);
      V1 := (Param - Single(Knots[I])) * Normalize(N - 1, I, Param);
      if D1 = 0 then
        V1 := 0
      else
        V1 := V1 / D1;
      D2 := Single(Knots[I + N + 1]) - Single(Knots[I + 1]);
      V2 := (Single(Knots[I + N + 1]) - Param) * Normalize(N - 1, I + 1, Param);
      if D2 = 0 then
        V2 := 0
      else
        V2 := V2 / D2;
      Result := V1 + V2;
    end;
  end;

  function GetNURBS(Index: Integer; Param: Double): TFPoint;
  var
    I: Integer;
    Ni: Double;
  begin
    Result.X := 0;
    Result.Y := 0;
    Result.Z := 0;
    for I := Index - 3 to Index do
    begin
      Ni := Normalize(3, I, Param);
      Result.X := Result.X + PFPoint(ControlPoints[I]).X * Ni;
      Result.Y := Result.Y + PFPoint(ControlPoints[I]).Y * Ni;
      Result.Z := Result.Z + PFPoint(ControlPoints[I]).Z * Ni;
    end;
  end;

  procedure AddPt(const Pt: TFPoint);
  var
    C: Integer;
    P: PInteger;
  begin
    C := Result.Count;
    Result.Count := Result.Count + cnstFPointSize;
    P := PInteger(Result.List);
    Inc(P,C);
    Move(Pt, P^, SizeOf(TFPoint));
  end;

var
  I: Integer;
  P, Prev: TFPoint;
  T: Double;
begin
  Result := TList.Create;
  P := PFPoint(ControlPoints[0])^;
  AddPt(P);
  Prev := P;
  for I := 3 to ControlPoints.Count - 1 do
  begin
    T := Single(Knots[I]);
    while  T < Single(Knots[I + 1]) do
    begin
      P := GetNURBS(I, T);
      T := T + 0.1; // the smaller value (for instance 0.01) the higher quality
      if (ChordLen > 0) and (abs(P.X - Prev.X) < ChordLen) and (abs(P.Y - Prev.Y) < ChordLen)
      then Continue;
      AddPt(P);
      Prev := P;
    end;
  end;
  P := PFPoint(ControlPoints[ControlPoints.Count - 1])^;
  AddPt(P);
end;

{ DrawSpline

  Draws a spline.  }
procedure DrawSpline;
var
  I: Integer;
  S: Single;
  ControlPoints, FitPoints, Knots, vList: TList;
  P, vP: PFPoint;
  SingPoints, IntPoints, DottedSingPoints: TList;
  PSinglePt: PFPoint;

  procedure DrawCurrentSpline;
  var I: Integer;
  begin
    if fmCADDLLdemo.cbUseWinLine.Checked then // Global coordinates
    begin
      if Lines.IsSolid then DrawGlobal(SingPoints, Lines.IsSolid)
      else begin
        Lines.Curve(SingPoints, DottedSingPoints, False);
        DrawGlobal(DottedSingPoints, False);
      end;
    end
    else                                   // Display coordinates
    begin
      IntPoints := TList.Create;
      if Lines.IsSolid then begin
        GetPointsList(SingPoints, IntPoints);
        DrawPolyLine(PCanvas.Handle, IntPoints);
      end
      else begin
        Lines.Curve(SingPoints, DottedSingPoints, False);
        GetPointsList(DottedSingPoints, IntPoints);
        DrawPolyPolyLine(PCanvas.Handle, IntPoints);
      end;
      for I := 0 to IntPoints.Count - 1 do Dispose(IntPoints[I]);
      IntPoints.Free;
    end;
  end;
begin
  SingPoints := TList.Create;
  DottedSingPoints := TList.Create;
  ControlPoints := TList.Create;
  FitPoints := TList.Create;
  Knots := TList.Create;
  try
    P := EData.Points;
    for I := 0 to EData.Count - 1 do begin
      New(vP);
      vP^ := P^;
      ControlPoints.Add(vP);
      {$IFDEF NOTDRAWSPLINE}
      New(PSinglePt);
      PSinglePt^.X := vP^.X;
      PSinglePt^.Y := vP^.Y;
      PSinglePt^.Z := vP^.Z;
      SingPoints.Add(PSinglePt);
      {$ENDIF}
      Inc(P);
    end;
    {$IFDEF NOTDRAWSPLINE}
    DrawCurrentSpline;
    Exit;
    {$ENDIF}
    for I := 0 to PInteger(P)^ - 1 do begin
      Inc(P);
      New(vP);
      vP^ := P^;
      FitPoints.Add(vP);
    end;
    Inc(P);
    for I := 0 to PInteger(P)^ - 1 do begin
      Inc(P);
      S := P^.X;
      Knots.Add(Pointer(S));
    end;
    if FitPoints.Count > ControlPoints.Count then begin
      vList := TList.Create;
      try
        for I := 0 to FitPoints.Count - 1 do begin
//          vList.Add(Pointer(PFPoint(FitPoints[I]).X));
//          vList.Add(Pointer(PFPoint(FitPoints[I]).Y));
//          vList.Add(Pointer(PFPoint(FitPoints[I]).Z));
          New(PSinglePt);
          PSinglePt^.X := PFPoint(FitPoints[I]).X;
          PSinglePt^.Y := PFPoint(FitPoints[I]).Y;
          PSinglePt^.Z := PFPoint(FitPoints[I]).Z;
          SingPoints.Add(PSinglePt);
        end;
        EData.Points := PFPoint(vList.List);
        EData.Count := FitPoints.Count;
        DrawCurrentSpline;
      finally
        vList.Free;
      end;
      Exit;
    end;
    vList := GetSpline(ControlPoints, FitPoints, Knots, 0);
    try
      EData.Points := PFPoint(vList.List);
      EData.Count := vList.Count div cnstFPointSize;
      P := EData.Points;
      for I := 0 to EData.Count-1 do begin
        New(PSinglePt);
        PSinglePt^.X := P^.X;
        PSinglePt^.Y := P^.Y;
        PSinglePt^.Z := P^.Z;
        SingPoints.Add(PSinglePt);
        Inc(P);
      end;
      DrawCurrentSpline;
    finally
      vList.Free;
    end;
  finally
    // Clearing memory
    for I := 0 to DottedSingPoints.Count - 1 do Dispose(DottedSingPoints[I]);
    for I := 0 to ControlPoints.Count - 1 do Dispose(ControlPoints[I]);
    for I := 0 to SingPoints.Count - 1 do Dispose(SingPoints[I]);
    for I := 0 to FitPoints.Count - 1 do Dispose(FitPoints[I]);
    DottedSingPoints.Free;
    ControlPoints.Free;
    SingPoints.Free;
    FitPoints.Free;
    Knots.Free;
  end;
end;

{ DrawText

  Draws a text. }
procedure DrawText;
var
  P: TPoint;
  vFontName: string;
  PPts: PFPoint;
  I, J, CntI, Cnt: Integer;
  IntPoints, Counts: TList;
  vPts: PInteger;
  vDC: HDC;

  procedure DoFontFromData;
  var
    I, Len: Integer;
    vLogFont: TLogFont;
  begin
    GetObject(PCanvas.Font.Handle, SizeOf(vLogFont), @vLogFont);
    FillChar(vLogFont.lfFaceName, LF_FACESIZE, 0);
    vFontName := string(EData.FontName);
    Len := Length(vFontName);
    if Len = 0 then
      vLogFont.lfFaceName := 'Arial'
    else// setting real TrueType font
    begin
      for I := 0 to LF_FACESIZE - 1 do
        if (I < Len) and (vFontName[I + 1] <> '.') then
          vLogFont.lfFaceName[I] := vFontName[I + 1]
        else
          Break;
    end;
    vLogFont.lfHeight := Round(fWinTextHeightFactor * EData.FHeight * fmCADDLLdemo.FScale / 100);
    if vLogFont.lfHeight = 0 then vLogFont.lfHeight := 1;
    vLogFont.lfWidth := Abs(Round(0.64 * EData.FHeight * EData.FScale * fmCADDLLdemo.FScale / 100));
    if vLogFont.lfWidth = 0 then vLogFont.lfWidth := 1;
    vLogFont.lfEscapement := Round(EData.Rotation * 10);	// in tenths of degrees
    while vLogFont.lfEscapement < 0 do
      Inc(vLogFont.lfEscapement, 3600);
    vLogFont.lfOrientation := vLogFont.lfEscapement;
    PCanvas.Font.Handle := CreateFontIndirect(vLogFont);
  end;

begin
  if bTextByCurves and (EData.DashDotsCount > 0) then
  begin
    IntPoints := TList.Create;
    Counts := TList.Create;
    try
      PPts := EData.DashDots;
      I := 0;
      while I < EData.DashDotsCount do
      begin
        Cnt := PInteger(PPts)^;
        Counts.Add(Pointer(Cnt));
        Inc(PPts);
        for J := 1 to Cnt do
        begin
          P := GetPoint(PPts^);
          IntPoints.Add(Pointer(P.X));
          IntPoints.Add(Pointer(P.Y));
          Inc(PPts);
        end;
        Inc(I);
      end;
      if EData.Flags and 1 <> 0 then
      begin
        // SHX text
        vDC := PCanvas.Handle;
        vPts := PInteger(IntPoints.List);
        CntI := Counts.Count - 1;
        for I := 0 to CntI do
        begin
          Cnt := Integer(Counts.List^[I]);
          if (Cnt = 2) and (vPts^ = PInteger(Integer(vPts) + 8)^) and
            (PInteger(Integer(vPts) + 4)^ = PInteger(Integer(vPts) + 12)^) then
            SetPixelV(vDC, vPts^, PInteger(Integer(vPts) + 4)^, PCanvas.Pen.Color)
          else
            Windows.Polyline(vDC, vPts^, Cnt);
          Inc(vPts, Cnt shl 1);
        end;
      end
      else
      begin
        // TTF text
        PCanvas.Brush.Style := bsSolid;
        PCanvas.Brush.Color := PCanvas.Pen.Color;
        Windows.PolyPolygon(PCanvas.Handle, IntPoints.List^, Counts.List^, Counts.Count);
        PCanvas.Brush.Style := bsClear;
      end;
    finally
      IntPoints.Free;
      Counts.Free;
    end;
  end
  else
  begin
    if EData.HAlign in [1,2,4] then
      P := GetPoint(EData.Point1)
    else
      P := GetPoint(EData.Point);
    DoFontFromData;
    PCanvas.Font.Color := PCanvas.Pen.Color;
    SetTextAlign(PCanvas.Handle, TA_BASELINE);
    PCanvas.Brush.Style := bsClear;
    PCanvas.TextOut(P.X, P.Y, EData.Text);
  end;
end;

{ DrawArcAsPoly

  Draws a arc as poplyline. }
procedure DrawArcAsPoly(const P0,P1,P2,P3: TFPoint);
const
  NSegs = 16; // Number of segments for full ellipse
var
  CX,CY,A,B,AStart,AEnd,Delta: Double;
  S,C: Extended;        // Sin and Cos of current angle
  I,N: Integer;
  P: TFPoint;
  SingPoints, IntPoints, DottedSingPoints: TList;
  PSinglePt: PFPoint;

  function Angle(const Pt: TFPoint): Extended;
  begin
    Result := ArcTan2(Pt.Y - CY, Pt.X - CX);
    if Result < 0 then Result := Result + Pi;
    if Pt.Y < CY then Result := Result + Pi;
  end;

begin
  SingPoints := TList.Create;
  DottedSingPoints := TList.Create;

  CX := (P0.X + P1.X) / 2;      // Center X
  CY := (P0.Y + P1.Y) / 2;      // Center Y
  A := (P1.X - P0.X) / 2;       // Horizontal radius
  B := (P0.Y - P1.Y) / 2;       // Vertical radius
  AStart := Angle(P2);          // Start angle
  AEnd := Angle(P3);            // End angle

  if AEnd <= AStart then AEnd := AEnd + 2*Pi;
  N := Round((AEnd-AStart)/Pi*NSegs);   // Real number of segments depends on arc angle
  if N < 4 then N := 4;
  Delta := (AEnd-AStart)/(N-1);

  for I:=0 to N-1 do
  begin
    SinCos(AStart,S,C);
    P.X := CX + A*C;                    // Current segment X
    P.Y := CY + B*S;                    // Current segment Y

    New(PSinglePt);
    PSinglePt^.X := P.X;
    PSinglePt^.Y := P.Y;
    PSinglePt^.Z := 0.0;// for future version
    SingPoints.Add(PSinglePt);
    AStart := AStart + Delta;           // Next segment
  end;

  if fmCADDLLdemo.cbUseWinLine.Checked then // Global coordinates
  begin
    if Lines.IsSolid then
      DrawGlobal(SingPoints, Lines.IsSolid)
    else
    begin
      Lines.Curve(SingPoints, DottedSingPoints, False);
      DrawGlobal(DottedSingPoints, Lines.IsSolid);
    end;
  end
  else
  begin
    IntPoints := TList.Create;
    if Lines.IsSolid then
    begin
      GetPointsList(SingPoints, IntPoints);
      DrawPolyLine(PCanvas.Handle, IntPoints);
    end
    else
    begin
      Lines.Curve(SingPoints, DottedSingPoints, False);
      GetPointsList(DottedSingPoints, IntPoints);
      DrawPolyPolyLine(PCanvas.Handle, IntPoints);
    end;
    for I := 0 to IntPoints.Count - 1 do
      Dispose(IntPoints[I]);
    IntPoints.Free;
  end;

  for I := 0 to DottedSingPoints.Count - 1 do
    Dispose(DottedSingPoints[I]);
  DottedSingPoints.Free;
  for I := 0 to SingPoints.Count - 1 do
    Dispose(SingPoints[I]);
  SingPoints.Free;
end;

{ DrawArc

  Draws a arc.
  EData.Point, EData.Point1 - bounding rectangle
  EData.Point2 - starting point
  EData.Point3 - ending point
  For full ellipse, Point2 = Point3.             }
procedure DrawArc;
var
  P,P1,P2,P3: TPoint;
  vRect: TRect;
  vRatio, vMajorLength: Single;
begin
  if fmCADDLLdemo.cbSplitArcs.Checked then
  begin
    DrawArcAsPoly(EData.Point, EData.Point1, EData.Point2, EData.Point3);
    Exit;
  end;
  P := GetPoint(EData.Point);
  P1 := GetPoint(EData.Point1);
  P2 := GetPoint(EData.Point2);
  P3 := GetPoint(EData.Point3);
  //Windows.SetArcDirection(PCanvas.Handle, AD_COUNTERCLOCKWISE);
  if ((EData^.Point2.X <> EData^.Point3.X)
    or (EData^.Point2.Y <> EData^.Point3.Y)) and (P2.X = P3.X)
    and (P2.Y = P3.Y) then
  begin
    SetPixelV(PCanvas.Handle, P2.X, P2.Y, PCanvas.Pen.Color);
    Exit;
  end;
  if (fmCADDLLdemo.cbProhibitArcsAsCurves.Checked) and (EData.EntityType=1) then
  begin
    vRatio := EData.Ratio;
    P := GetPoint(EData.Point);
    P1 := GetPoint(MakeFPoint(EData.Point.X + EData.Point1.X, EData.Point.Y + EData.Point1.Y, 0));
    vMajorLength := Sqrt(Sqr(P1.X - P.X) + Sqr(P1.Y - P.Y));
    vRect.Left := P.X - Round(vMajorLength);
    vRect.Top := P.Y + Round(vMajorLength * vRatio);
    vRect.Right := P.X + Round(vMajorLength);
    vRect.Bottom := P.Y - Round(vMajorLength * vRatio);
    PCanvas.Arc(vRect.Left, vRect.Top, vRect.Right, vRect.Bottom, P2.X, P2.Y, P3.X, P3.Y);
  end
  else
    if (EData.StartAngle = EData.EndAngle) then
      SetPixelV(PCanvas.Handle, P2.X, P2.Y, PCanvas.Pen.Color)
    else
      PCanvas.Arc(P.X, P1.Y, P1.X, P.Y, P2.X, P2.Y, P3.X, P3.Y);
end;

{ DrawHatch

  Draws a hatch.
  EData.Count - number of polyline vertices
  EData.Points - pointer to point array }
procedure DrawHatch;
var
  I, SaveIndex, Index: Integer;
  P: PFPoint;
  PSinglePt: PFPoint;
  vPoly, SingPoints: TList;
  vRgn, vMainRGN: HRGN;
  P1: TPoint;
begin
  if EData.Flags and 16 <> 0 then             // hatch is SOLID
  begin
    vPoly := TList.Create;
    vMainRGN := CreateRectRgn(0, 0, 0, 0);// FCanvas.ClipRect.Right, FCanvas.ClipRect.Bottom);
    SaveIndex := SaveDC(PCanvas.Handle);
    try
      P := EData.Points;
      Index := 0;// this is a counter of points for each boundary
      I := 0;    // this is a counter of boundaries themselves
      while I <= EData.Count do
      begin
        if Index = 0 then
        begin
          if I <> 0 then
          begin
            vRGN := CreatePolygonRgn(vPoly.List^, vPoly.Count shr 1, ALTERNATE);
            CombineRgn(vMainRGN, vRGN, vMainRGN,  RGN_XOR);
            DeleteObject(vRGN);
            vPoly.Clear;
            if I = EData.Count then
              Break;
          end;
          Index := PInteger(P)^;
          Inc(I);
        end
        else
        begin
          Dec(Index);
          P1 := GetPoint(P^);
          vPoly.Add(Pointer(P1.X));
          vPoly.Add(Pointer(P1.Y));
        end;
        Inc(P);
      end;
      GetRegionData(vMainRgn,0,nil);
      PCanvas.Brush.Style := bsSolid;
      PCanvas.Brush.Color := EData.Color;
      FillRgn(PCanvas.Handle, vMainRGN, PCanvas.Brush.Handle);
    finally
      RestoreDC(PCanvas.Handle, SaveIndex);
      DeleteObject(vMainRGN);
      vPoly.Free;
    end;
    Exit;
  end;
  SingPoints := TList.Create;
  P := EData.DashDots;
  for I := 0 to EData.DashDotsCount - 1 do
  begin
    New(PSinglePt);
    PSinglePt^.X := P^.X;
    PSinglePt^.Y := P^.Y;
    PSinglePt^.Z := P^.Z;
    SingPoints.Add(PSinglePt);
    Inc(P);
  end;
  DrawGlobal(SingPoints, False);
  // Clearing memory
  for I := 0 to SingPoints.Count - 1 do
    Dispose(SingPoints[I]);
  SingPoints.Free;
end;

{ DrawImageEnt

  Draw a bmp image}
procedure DrawImageEnt;
var
  P1,P2: TPoint;
//  P,P3: TPoint;
  bmp: TBitmap;
  vStr: TMemoryStream;
  I: Integer;
  vRect: TRect;
begin
//  P := GetPoint(EData.Point);
  P1 := GetPoint(EData.Point1);
  P2 := GetPoint(EData.Point2);
//  P3 := GetPoint(EData.Point3);
  vRect.TopLeft := P2;
  vRect.BottomRight := P1;
  if vRect.Right < vRect.Left then
  begin
    I := vRect.Right;
    vRect.Right := vRect.Left;
    vRect.Left := I;
  end;
  if vRect.Bottom < vRect.Top then
  begin
    I := vRect.Bottom;
    vRect.Bottom := vRect.Top;
    vRect.Top := I;
  end;
  vStr := TMemoryStream.Create;
  bmp := TBitmap.Create;
  try
    vStr.Write(EData.Ticks^, EData.Handle);
    vStr.Position := 0;
    bmp.LoadFromStream(vStr);
    PCanvas.StretchDraw(vRect, bmp);
  finally
    vStr.Free;
    bmp.Free;
  end;
end;

{ BeginViewport

  Creates a clipping region according to the VIEWPORT's boundary.
  Makes necessary actions before drawing the VIEWPORT and his "contents".    }
procedure BeginViewport;
var
  L: TList;
  P: PFPoint;
  TP: TPoint;
  I, J, K: Integer;
  MainRgn, Rgn: HRgn;
  R: TRect;
begin
  R.TopLeft := GetPoint(EData.Point);
  R.BottomRight := GetPoint(EData.Point1);
  if R.Left > R.Right then SwapInts(R.Left, R.Right);
  if R.Top > R.Bottom then SwapInts(R.Top, R.Bottom);
  SaveDC(PCanvas.Handle);
  if EData.Count = 0 then
  begin
    MainRgn := CreateRectRgnIndirect(R);
    if EData.Flags and 1 <> 0 then
      PCanvas.Rectangle(R.Left, R.Top, R.Right, R.Bottom);
  end
  else
  begin
    P := EData.Points;
    I := EData.Count;
    MainRgn := CreateRectRgn(0,0,0,0);
    L := TList.Create;
    try
      while I > 0 do
      begin
        L.Count := 0;
        K := PInteger(P)^;
        Inc(P);
        for J := 0 to K - 1 do
        begin
          TP := GetPoint(P^);
          Inc(P);
          L.Add(Pointer(TP.X));
          L.Add(Pointer(TP.Y));
        end;
        Rgn := CreatePolygonRgn(L.List^, K, ALTERNATE);
        CombineRgn(MainRgn, MainRgn, Rgn, RGN_XOR);
        DeleteObject(Rgn);
        Dec(I);
      end;
    finally
      L.Free;
    end;
  end;
  TP := Point(0,0);
  LPtoDP(PCanvas.Handle, TP, 1);
  OffsetRgn(MainRgn, TP.X, TP.Y);
  SelectClipRgn(PCanvas.Handle, MainRgn);
  DeleteObject(MainRgn);
end;

{ EndViewport

  Makes necessary actions after drawing the VIEWPORT and his "contents".    }
procedure EndViewport;
begin
  RestoreDC(PCanvas.Handle, -1);
end;

{ DoLines

  Creates the list of ticks. }
procedure DoLines;
type
  PSingle = ^Single;
var
  I: Integer;
  P: PSingle;
  vPElement: PsgLTypeElement;
begin
  Lines.Clear;
  Lines.Scale := EData.Rotation*LScale;
  P := PSingle(EData.Ticks);
  for I := 0 to EData.TickCount - 1 do
  begin
    New(vPElement);
    FillChar(vPElement^, SizeOf(vPElement^), 0);
    vPElement^.Dash := P^;
    Lines.Add(vPElement);
    Inc(P);
  end;
  Lines.Loaded(nil);
end;

{ DoDraw

  Called from DXFEnum for each entity in CADHandle file
  Data points to entity data
  Param is user-defined parameter. }
procedure DoDraw(Data: PcadData; var Param); stdcall;
begin
  // if DXFVisible(CADHandle, Data.Layer) = 0 then
  //   Exit; // invisible layers are ignored
  // MessageBox(0, PChar(Format('%d', [SizeOf(Data)])), 'a', 0);
  if not fmCADDLLdemo.cbDimensions.Checked and (Data.Dimension <> 0) then Exit;
  EData := Data;
  PCanvas.Pen.Color := Data.Color;      // entity color
  PCanvas.Pen.Width := Round(Data.Thickness * fmCADDLLdemo.FScale / 100);
  DoLines;
  case EData.Tag of
    DXF_LINE:                           DrawLine;
    DXF_SOLID:          	              DrawSolid;
    DXF_3DFACE:                         Draw3DFace;
    DXF_CIRCLE,DXF_ARC,DXF_ELLIPSE:     DrawArc;
    DXF_POLYLINE..DXF_LWPOLYLINE:       DrawPoly;
    DXF_SPLINE:                         DrawSpline;
    DXF_TEXT, DXF_ATTDEF, DXF_ATTRIB:   DrawText;
    DXF_POINT:		                      DrawPoint;
    DXF_HATCH:                          DrawHatch;
    DXF_IMAGE_ENT:                      DrawImageEnt;
    DXF_BEGIN_INSERT:                   IsInsideInsert := True;
    DXF_END_INSERT:                     IsInsideInsert := False;
    DXF_BEGIN_VIEWPORT:                 BeginViewport;
    DXF_END_VIEWPORT:                   EndViewport;
  end;
end;

{ TfmCADDLLdemo.pbDrawingPaint

  Use bit-coded DXFEnum flag =4 for setting a way of cutting a viewport's
  contents to fit a "rectangular" VIEWPORT, without using Windows API REGIONS.
  If you have no possibility to use Windows API REGIONS
  (for example, you use OpenGL or something similar), set this flag.  }
procedure TfmCADDLLdemo.pbDrawingPaint(Sender: TObject);
begin
  if CADHandle <> 0 then
  begin
    Lines.Scale := LScale;
    pbDrawing.Canvas.Brush.Color := Color;
    pbDrawing.Canvas.FillRect(pbDrawing.Canvas.ClipRect);
    CADEnum(CADHandle, Ord(sbAllLayers.Checked) or  (Integer(bTextByCurves) shl 3), @DoDraw, Self);        // use DXFEnum for import as a linear "metafile"
  end;
end;

procedure TfmCADDLLdemo.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if CADHandle <> 0 then
    CADClose(CADHandle);
end;

procedure TfmCADDLLdemo.cbScaleChange(Sender: TObject);
begin
  FScale := StrToInt(cbScale.Text);
  pbDrawing.Invalidate;
end;

procedure TfmCADDLLdemo.pbDrawingMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbRight then
  begin
    FStart := Point(X,Y);
    FOld := Point(FX,FY);
    pbDrawing.Cursor := crHandPoint;
    TMyPaint(pbDrawing).MouseCapture := True;
    tsDrawing.Perform(WM_SETCURSOR, tsDrawing.Handle, HTCLIENT);
  end;
end;

procedure TfmCADDLLdemo.pbDrawingMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if pbDrawing.Cursor = crHandPoint then
  begin
    FX := FOld.X + X - FStart.X;
    FY := FOld.Y + Y - FStart.Y;
    FPlaceToHome := False;
    pbDrawing.Invalidate;
  end;
end;

procedure TfmCADDLLdemo.pbDrawingMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if (Button = mbRight) and (pbDrawing.Cursor = crHandPoint) then
  begin
    TMyPaint(pbDrawing).MouseCapture := False;
    pbDrawing.Cursor := crDefault;
  end;
end;

procedure TfmCADDLLdemo.cbLayersDrawItem(Control: TWinControl; Index: Integer;
  Rect: TRect; State: TOwnerDrawState);
var C: TColor;
begin
  C := TColor(cbLayers.Items.Objects[Index]);
  with cbLayers.Canvas do
  begin
    Brush.Color := clWindow;
    FillRect(Rect);
    if C < 0 then Font.Color := clGray else Font.Color := clBlack;
    TextRect(Rect, Rect.Left+16, Rect.Top, cbLayers.Items[Index]);
    Brush.Color := C and MaxInt;
    Rectangle(Rect.Left+2, Rect.Top+2, Rect.Left+12, Rect.Top+12);
  end;
end;

procedure TfmCADDLLdemo.FormCreate(Sender: TObject);
begin
  Lines := TsgLines.Create;
  FPlaceToHome := False;
  FZoomAll := False;
  FX := 0;
  FY := 0;
  btnPlaceImageToHome.Enabled:= False;
  cbTextAsCurves.Checked := True;
end;

procedure TfmCADDLLdemo.FormDestroy(Sender: TObject);
begin
  Lines.Free;
end;

procedure TfmCADDLLdemo.mmiOpenClick(Sender: TObject);
const S: array[0..1] of string = ('Inches','Millimeters');
var
  I, Cnt: Integer;
  C: Cardinal;
  Data: TcadData;
  vLayoutName: string;
begin
  try
    if mmiOpen.Checked then
      Exit;
    OpenDialog1.InitialDir := FCurrentDir;
    if not OpenDialog1.Execute then
      Exit;
    mmiOpen.Checked := True;
    if Length(OpenDialog1.FileName) > 1 then
      FCurrentDir := ExtractFilePath(OpenDialog1.FileName);
    if CADHandle <> 0 then CADClose(CADHandle);
    CADHandle := CADCreate(Application.Handle, PChar(OpenDialog1.FileName));    // Open CADHandle file
    if CADHandle = 0 then Error;
    if CADUnits(CADHandle,I) = 0 then Error;    // Drawing units (inches or millimeters)
    if CADLTScale(CADHandle,LScale) = 0 then Error;
    lblMeasure.Caption := S[I];
    FScale := 100;
    cbScale.ItemIndex := 7;
    cbLayers.Items.Clear;
    Cnt := CADLayerCount(CADHandle);
    for I := 0 to Cnt - 1 do
    begin
      CADLayer(CADHandle,I,@Data);
{  Data.flags bit code
1 - Visible
2 - Frozen
4 - Locked
8 - IsPlotting
}
      C := Data.Color;
      if Data.Flags and 3 <> 1 then
        C := C or $80000000;
        
      cbLayers.Items.AddObject(Data.Text, TObject(C));
    end;
    cbLayouts.Items.Clear;
    Cnt := CADLayoutCount(CADHandle);
    for I := 0 to Cnt - 1 do
    begin
      SetLength(vLayoutName, 100);
      CADLayoutName(CADHandle, I, PChar(vLayoutName), 100);
      cbLayouts.Items.Add(vLayoutName);
    end;
    if CADLayoutCurrent(CADHandle, DWORD(I), False) = 1 then
      cbLayouts.ItemIndex := I;
    FX := 0;
    FY := 0;
    FZoomAll := True;
    mmiPlaceImageToHomeClick(nil);
    cbProhibitArcsAsCurvesClick(nil);
    btnPlaceImageToHome.Enabled := True;
    cbScale.SetFocus;
    mmiOpen.Checked := False;
  except
    mmiOpen.Checked := False;
  end;
end;

procedure TfmCADDLLdemo.mmiPlaceImageToHomeClick(Sender: TObject);
begin
  FPlaceToHome := True;
  mmiZoomAllClick(nil);
end;

procedure TfmCADDLLdemo.Exit1Click(Sender: TObject);
begin
  Close;
end;

procedure TfmCADDLLdemo.cbSplitArcsClick(Sender: TObject);
begin
  pbDrawing.Invalidate;
end;

procedure TfmCADDLLdemo.cbTextAsCurvesClick(Sender: TObject);
begin
  bTextByCurves := cbTextAsCurves.Checked;
  pbDrawing.Invalidate;
end;

procedure TfmCADDLLdemo.cbUseWinLineClick(Sender: TObject);
begin
  pbDrawing.Invalidate;
end;

procedure TfmCADDLLdemo.cbProhibitArcsAsCurvesClick(Sender: TObject);
begin
  // 1 => permit conversion of arcs to polyline
  CADProhibitCurvesAsPoly(CADHandle, Ord(not fmCADDLLdemo.cbProhibitArcsAsCurves.Checked));
  pbDrawing.Invalidate;
end;

procedure TfmCADDLLdemo.pmiShowAboutClick(Sender: TObject);
begin
  ShowMessage('CAD Importer SDK DLL Version.' + #13 +
    'www.cadsofttools.com');
end;

procedure TfmCADDLLdemo.mmiZoomAllClick(Sender: TObject);
var
  L, T, R, B, W, H: Double;
  vNewScaleStr: string;
begin
  if CADGetBox(CADHandle, L, R, T, B) = 0 then Exit;
  W := R - L;
  if W = 0 then
    W := 1;
  H := T - B;
  if H = 0 then
    H := 1;
  W := pbDrawing.Width * 100 / W;
  H := pbDrawing.Height * 100 / H;
  if W > H then
    W := H;
  if (FZoomAll = True) or (Sender <> nil) then
  begin
    FScale := Round(W);
    vNewScaleStr := IntToStr(FScale);
    if cbScale.Items.IndexOf(vNewScaleStr) < 0 then
      cbScale.Items.AddObject(vNewScaleStr, nil);
    cbScale.ItemIndex := cbScale.Items.IndexOf(vNewScaleStr);
    FZoomAll := False;
  end;
  if FPlaceToHome = True then
  begin
    FX := - Round(L * FScale / 100); // see GetPoint
    FY := Round(T * FScale / 100);
  end;
  pbDrawing.Invalidate;
end;

procedure TfmCADDLLdemo.cbLayoutsChange(Sender: TObject);
var
  I: DWORD;
begin
  if CADHandle <> 0 then
  begin
    I := cbLayouts.ItemIndex;
    CADLayoutCurrent(CADHandle, I, True);
    FPlaceToHome := True;
    mmiZoomAllClick(Sender);
    cbScale.SetFocus;
  end;
end;

initialization
{$IFDEF MEMCHK}
  MemChk
{$ENDIF}

end.
